from typing import Callable, Type, TypeVar

from fastapi_global_variable import GlobalVariable
from fastapi_request_helper.constants.http_method import POST
from fastapi_request_helper.decorators.controller import add_function_routers
from fastapi_request_helper.helpers.router_decorator import set_api_detail
from fastapi_utils.cbv import cbv
from fastapi_utils.inferring_router import InferringRouter

T = TypeVar('T')


def action():
    def decorator(cls: Type[T]):
        set_api_detail(cls, 'methods', [POST])
        set_api_detail(cls, 'path', '/' + cls.__name__)
        return cls

    return decorator


def microservice() -> Callable[[Type[T]], Type[T]]:
    def decorator(cls: Type[T]) -> Type[T]:
        router = InferringRouter()
        prefix = f'microservices/{cls.__name__}'
        add_function_routers(prefix, cls, router)
        cbv(router)(cls)
        GlobalVariable.get_or_fail('app').include_router(router)

        return cls

    return decorator



from types import GenericAlias


def oneline(value: str):
    return value + '\r\n'


def tab(content: str, tabLength=1):
    tabContent = '    ' * tabLength
    newContent = ''
    for line in content.split('\r\n'):
        newContent += oneline(tabContent + line)

    return newContent


def get_generic_type(ref_type: type(GenericAlias)):
    alias_params = map(lambda ref: ref.__name__, ref_type.__args__)
    return f'{ref_type.__name__}[{",".join(alias_params)}]'



from dataclasses import dataclass

from fastapi.encoders import jsonable_encoder
from requests import Session, Request


@dataclass
class MicroserviceOption:
    is_json: bool = True
    headers: dict = None


class BaseMicroserviceClient:
    http_client: Session = Session()

    @staticmethod
    def send(url: str, query_params: dict, body_params: any, option: MicroserviceOption = None):
        if not option:
            option = MicroserviceOption()
        request_config = {
            'method': 'POST',
            'url': url,
            'headers': option.headers,
            'params': query_params,
            'data': body_params if not option.is_json else None,
            'json': jsonable_encoder(body_params) if option.is_json else None,
        }

        prepped_request = BaseMicroserviceClient.http_client.prepare_request(Request(**request_config))
        return BaseMicroserviceClient.http_client.send(prepped_request)



from .decorators.controller import controller
from .decorators.http_method import (delete, description, get, guard, guards, hidden_when, name, patch, post, put,
                                     rate_limit, read_only, response, status, status_created, status_no_content,
                                     status_ok, summary, tag)
from .decorators.authentication import auth, current_user_id
from .dependencies.pagination import PaginationParams

__all__ = (
    'controller',
    'get',
    'post',
    'put',
    'delete',
    'patch',
    'response',
    'guard',
    'guards',
    'rate_limit',
    'status',
    'description',
    'summary',
    'name',
    'tag',
    'status_no_content',
    'read_only',
    'status_ok',
    'status_created',
    'hidden_when',
    'PaginationParams',
    'auth',
    'current_user_id'
)



import dataclasses
from dataclasses import dataclass, is_dataclass
from inspect import getsourcefile
from types import GenericAlias
from typing import Dict

import pydash
from fastapi.routing import APIRoute
from pydantic.fields import ModelField
from pydantic.main import ModelMetaclass
from pydash import group_by, some
from ssort import ssort
from starlette.routing import BaseRoute

from .base_microservice_client import BaseMicroserviceClient
from .code_generation import GenerationPropertyDetail, GenerationClassDetail, Generation, \
    GenerationMethodDetail, CloneGeneration, GenerationImportModuleDetail
from .helper import oneline


@dataclass
class MethodDetail:
    class_name: str
    path: str
    function_name: str
    query_params: list[ModelField]
    body_params: list[ModelField]
    parameters_position: list
    response_field: ModelField


class SdkTemplate:
    def generate_method(self, method_detail: MethodDetail):
        params: list[ModelField] = list(method_detail.query_params + method_detail.body_params)
        params.sort(key=lambda param: method_detail.parameters_position.index(param.name))

        return GenerationMethodDetail(
            name=method_detail.function_name,
            response=self.get_response(method_detail),
            body=self.get_method_body(method_detail),
            params=list(map(lambda param: GenerationPropertyDetail(param.name, param.type_.__name__), params))
        )

    def get_method_body(self, method_detail: MethodDetail):
        body_params = [f"'{method_detail.path}'"]

        if method_detail.query_params:
            query_params = ', '.join(map(lambda query: f"'{query.name}': {query.name}", method_detail.query_params))
            body_params.append('{' + query_params + '}')
        else:
            body_params.append('None')

        if method_detail.body_params:
            body_params.append(method_detail.body_params[0].name)
        else:
            body_params.append('None')

        return f'return self.send({", ".join(body_params)}, option )'

    def get_response(self, method_detail: MethodDetail):
        if not method_detail.response_field:
            return 'None'
        response_name = method_detail.response_field.type_.__name__
        if isinstance(method_detail.response_field.annotation, GenericAlias):
            return f'{method_detail.response_field.annotation.__name__}[{response_name}]'
        return response_name

    def generate_class(self, name, methods: list[MethodDetail]):
        class_detail = GenerationClassDetail(class_name=name, parent_name='BaseMicroserviceClient')
        for method in methods:
            class_detail.methods.append(self.generate_method(method))
        return Generation().generate(class_detail)

    def generate_ref_class(self, microservice: Dict):
        list_ref_class = list()
        list_import_module: list[GenerationImportModuleDetail] = list((
            GenerationImportModuleDetail('dataclasses', 'dataclass'),
            GenerationImportModuleDetail('dataclasses', 'field'),
            GenerationImportModuleDetail('enum', 'Enum')
        ))
        for key in microservice.keys():
            methods: list[MethodDetail] = microservice[key]
            for method in methods:
                for body in method.body_params:
                    self.get_list_ref_class(body.type_, list_ref_class, list_import_module)
                for query in method.query_params:
                    self.get_list_ref_class(query.type_, list_ref_class, list_import_module)
                self.get_list_ref_class(method.response_field.type_, list_ref_class, list_import_module)

        content = oneline(Generation().generate_import_modules(list_import_module))
        list_ref_class = pydash.filter_(list_ref_class, lambda ref_class: not isinstance(ref_class, GenericAlias))

        for ref_class in list_ref_class:
            content += CloneGeneration().clone_class_to_dataclass(ref_class)
        return content

    def get_list_ref_class(self, ref_type, list_ref_class: list,
                           list_import_module: list[GenerationImportModuleDetail]):
        if some(list_import_module,
                lambda import_module: import_module.type == ref_type.__qualname__ and import_module.module == ref_type.__module__):
            return
        if type(ref_type) is type and not is_dataclass(ref_type):
            print(ref_type)
            if ref_type.__module__ != 'builtins':
                list_import_module.append(GenerationImportModuleDetail(ref_type.__module__, ref_type.__qualname__))
            return
        if ref_type in list_ref_class:
            return
        list_ref_class.append(ref_type)

        if type(ref_type) is ModelMetaclass:
            for key in ref_type.__fields__.keys():
                property: ModelField = ref_type.__fields__[key]
                self.get_list_ref_class(property.type_, list_ref_class, list_import_module)
        elif is_dataclass(ref_type):
            for field in dataclasses.fields(ref_type):
                self.get_list_ref_class(field.type, list_ref_class, list_import_module)
        elif isinstance(ref_type, GenericAlias):
            for alias in ref_type.__args__:
                self.get_list_ref_class(alias, list_ref_class, list_import_module)


class SdkBuilder:
    def generate(self, routes: list[BaseRoute]) -> None:
        microservices = list()
        for route in routes:
            if isinstance(route, APIRoute):
                if route.path.startswith('/microservices/'):
                    microservices.append(self.get_api_detail(route))

        microservice = group_by(microservices, 'class_name')
        self.create_file(microservice)

    def get_api_detail(self, route: APIRoute):
        qualified_name: str = route.endpoint.__qualname__
        parameters_position = list(route.endpoint.__code__.co_varnames)
        parameters_position.remove('self')

        if len(route.dependant.body_params) > 1:
            raise Exception('Cannot use more than 1 dto for microservices.')

        method_detail = MethodDetail(
            path=route.path,
            class_name=qualified_name.split('.')[0],
            function_name=qualified_name.split('.')[1],
            query_params=route.dependant.query_params,
            body_params=route.dependant.body_params,
            parameters_position=parameters_position,
            response_field=route.response_field
        )
        return method_detail

    def create_file(self, microservice: Dict):
        sdk_template = SdkTemplate()
        content = ''''''
        content += sdk_template.generate_ref_class(microservice)
        for key in microservice.keys():
            content += sdk_template.generate_class(key, microservice[key])
        with open(getsourcefile(BaseMicroserviceClient), 'r') as file:
            content += file.read()
        print(ssort(content))

        # with open('test.py', "w") as file:
        #     file.write(content)



import dataclasses
import typing
from dataclasses import field, dataclass, is_dataclass, _MISSING_TYPE
from enum import Enum
from types import GenericAlias

from pydantic.fields import ModelField
from pydantic.main import ModelMetaclass
from pydash import group_by

from .helper import oneline, tab, get_generic_type


@dataclass
class GenerationImportModuleDetail:
    module: str
    type: str


@dataclass
class GenerationPropertyDetail:
    name: str
    type: typing.Optional[str] = None
    default_value: typing.Optional[any] = None
    default_factory: typing.Optional[any] = None
    allow_none: bool = False


@dataclass
class GenerationMethodDetail:
    name: str
    params: list[GenerationPropertyDetail] = field(default_factory=list)
    body: str = None
    response: str = None


@dataclass
class GenerationClassDetail:
    class_name: str
    parent_name: str = None
    methods: list[GenerationMethodDetail] = field(default_factory=list)
    properties: list[GenerationPropertyDetail] = field(default_factory=list)
    is_dataclass: bool = False


class Generation:
    def generate(self, class_detail: GenerationClassDetail):
        return self.generate_class(class_detail)

    def generate_import_modules(self, list_import_module: list[GenerationImportModuleDetail]):
        content = ''
        import_modules = group_by(list_import_module, 'module')
        for key in import_modules.keys():
            types = map(lambda import_module: import_module.type, import_modules[key])
            content += oneline(f'from {key} import {",".join(types)}')
        return content

    def generate_class(self, class_detail: GenerationClassDetail):
        parent_name = f'({class_detail.parent_name})' if class_detail.parent_name else ''
        content = oneline('@dataclass') if class_detail.is_dataclass else ''
        content += oneline(f'class {class_detail.class_name}{parent_name}:')
        content += self.generate_properties(class_detail.properties)
        content += self.generate_methods(class_detail.methods)
        return content

    def generate_methods(self, methods: list[GenerationMethodDetail]):
        content = ''
        for method in methods:
            content += tab(self.generate_method(method))
            content += '\r\n'
        return content

    def generate_method(self, method: GenerationMethodDetail):
        paramList = ", ".join(map(lambda param: f'{param.name}: {param.type}', method.params))
        content = oneline(
            f'def {method.name}(self, {paramList}, option: MicroserviceOption = None) -> {method.response}:')
        content += oneline(tab(method.body))

        return content

    def generate_properties(self, properties: list[GenerationPropertyDetail]):
        content = ''
        for property in properties:
            content += tab(self.generate_property(property))
            content += '\r\n'
        return content

    def generate_property(self, property: GenerationPropertyDetail):
        content = f'{property.name}: {property.type}' if property.type else property.name
        if property.default_value:
            content += f'={property.default_value}'
        if property.default_value is None and property.allow_none:
            content += f'=None'
        elif property.default_factory:
            content += f'=field(default_factory={property.default_factory})'
        return content


class CloneGeneration:
    def clone_class_to_dataclass(self, ref_class):
        if type(ref_class) is ModelMetaclass:
            class_detail = GenerationClassDetail(class_name=ref_class.__name__, is_dataclass=True)
            for key in ref_class.__fields__.keys():
                property: ModelField = ref_class.__fields__[key]
                type_name = property.type_.__name__
                if isinstance(property.annotation, GenericAlias):
                    type_name = get_generic_type(property.annotation)
                property_detail = GenerationPropertyDetail(
                    property.name,
                    type=type_name,
                    default_value=property.default,
                    default_factory=property.default_factory,
                    allow_none=property.allow_none
                )
                class_detail.properties.append(property_detail)

            return Generation().generate(class_detail)
        else:
            if issubclass(ref_class, Enum):
                return self.clone_enum(ref_class)
            elif is_dataclass(ref_class):
                return self.clone_dataclass(ref_class)
            else:
                print(type(ref_class))
                print("Need to implement this type", ref_class)

        return ''

    def clone_enum(self, ref_class: type[Enum]):
        class_detail = GenerationClassDetail(class_name=ref_class.__name__, parent_name=Enum.__name__)
        for member in ref_class:
            class_detail.properties.append(GenerationPropertyDetail(
                name=member.name,
                default_value=member.value
            ))
        return Generation().generate(class_detail)

    def clone_dataclass(self, ref_class: type[dataclasses]):
        class_detail = GenerationClassDetail(class_name=ref_class.__name__,
                                             is_dataclass=True)
        for field in dataclasses.fields(ref_class):
            type_name = field.type.__name__
            if isinstance(field.type, GenericAlias):
                type_name = get_generic_type(field.type)
            property_detail = GenerationPropertyDetail(
                name=field.name,
                type=type_name,
                default_value=field.default if not type(field.default) is _MISSING_TYPE else None,
                default_factory=field.default_factory if not type(field.default_factory) is _MISSING_TYPE else None
            )
            class_detail.properties.append(property_detail)
        return Generation().generate(class_detail)
